#include <exec/types.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include "locale_intern.h"
#include "private.h"
#include <exec/execbase.h>

void dprintf(const char *, ...);

#undef SysBase
#undef DOSBase
extern struct ExecBase *SysBase;
extern struct DosLibrary *DOSBase;

extern struct SignalSemaphore timezonelock;
extern timezone_t systemtimezone;

struct PrivateTimerDevBase
{
	struct Device dev;
	BPTR bptr1;
	APTR ptr1;
	APTR ptr2;
	struct Library *LocaleBase; /* guaranteed offset with V52+ timer.device */
};

struct PrivateBattClockResBase
{
	struct Library lib;
	UWORD pad1;
	APTR ptr1;
	APTR ptr2;
	struct Library *LocaleBase; /* guaranteed offset with V52+ battclock.resource */
};

void install_timezone_patches(struct LocaleBase *LocaleBase)
{
	struct PrivateTimerDevBase *TimerBase;
	struct PrivateBattClockResBase *BattClockRes;

	Disable();

	TimerBase = (APTR) FindExecNode(EXECLIST_DEVICE, "timer.device");
	BattClockRes = (APTR) FindExecNode(EXECLIST_RESOURCE, "battclock.resource");

	if (TimerBase && TimerBase->dev.dd_Library.lib_Version >= 52)
	{
		TimerBase->LocaleBase = (APTR) LocaleBase;
	}
	if (BattClockRes && BattClockRes->lib.lib_Version >= 52)
	{
		BattClockRes->LocaleBase = (APTR) LocaleBase;
	}

	Enable();
}

ULONG TZGetSysLocalTime2(ULONG _utctime, CONST_STRPTR timezone, struct tm *restm, struct LocaleBase *LocaleBase)
{
  time_t utctime = (time_t) _utctime;
  struct tm tmp;
  time_t rc;

  if (timezone)
  {
    /* This codepath is not safe from situations where FS might trigger looping */
    timezone_t tz = tzalloc(timezone);
    if (!tz)
      return ~0;

    if (localtime_rz(tz, &utctime, &tmp) == NULL)
    {
      tzfree(tz);
      return ~0;
    }

    if (restm)
      *restm = tmp;
    tmp.tm_isdst = 0;
    rc = mktime_z(NULL, &tmp);

    tzfree(tz);
  }
  else
  {
    /*
       This codepath attempts to be safe from semaphore lockups / recursion
       issues:
       1) It uses Forbid() to make it less b0rk when called from timer.device etc.
       2) Everything should be preloaded to memory at this stage so that no
          Wait() will be called or recursion will occurt from loading resources
          from a file system.

       Yes, it is a bit delicate and depends on the tzdb internals. - Piru
    */
    Forbid();
    //ObtainSemaphoreShared(&timezonelock);

    //dprintf("%s: convert unix ts %u to local timestamp\n", __func__, utctime);

    if (systemtimezone == 0)
    {
      //ReleaseSemaphore(&timezonelock);
      //dprintf("%s: systemtimezone is NULL, return ~0\n", __func__);
      Permit();
      return ~0;
    }

    if (localtime_rz(systemtimezone, &utctime, &tmp) == NULL)
    {
      //ReleaseSemaphore(&timezonelock);
      //dprintf("%s: localtime_rz failed, return ~0\n", __func__);
      Permit();
      return ~0;
    }

    if (0) {
      char buf[40];
      asctime_r(&tmp, buf); buf[strlen(buf) - 1] = '\0';
      dprintf("%s: local time is %d.%d.%d %02d:%02d:%02d tm_isdst %d (%s)\n", __func__, tmp.tm_mday, tmp.tm_mon + 1, tmp.tm_year + 1900, tmp.tm_hour, tmp.tm_min, tmp.tm_sec, tmp.tm_isdst, buf);
    }
    if (restm)
      *restm = tmp;
    tmp.tm_isdst = 0;
    /*
       This code relies on TZSetSysTimeZone having preloaded the GMT zoneinfo.
       If this not done, it could result in a infinite recursion if the filesystem
       would call GetSysTime() or TR_GETSYSTIME (which two in turn call this
       function). - Piru
     */
    rc = mktime_z(NULL, &tmp);
    //dprintf("%s: resulting local timestamp %u\n", __func__, rc);

    //ReleaseSemaphore(&timezonelock);
    Permit();
  }

  /*
     Sanitize the return value.
   */
  if (rc < 0 || rc > 0xffffffffLL)
    rc = ~0;

  return (ULONG) rc;
}

ULONG TZGetSysLocalTime(ULONG utctime, CONST_STRPTR timezone, struct LocaleBase *LocaleBase)
{
  return TZGetSysLocalTime2(utctime, timezone, NULL, LocaleBase);
}

